#ifndef PBCOPPER_CLI_v2_HIDDENOPTIONMODE_H
#define PBCOPPER_CLI_v2_HIDDENOPTIONMODE_H

#include <pbcopper/PbcopperConfig.h>

namespace PacBio {
namespace CLI_v2 {
namespace internal {

enum class HiddenOptionMode
{
    HIDE,
    SHOW,
};

}  // namespace internal
}  // namespace CLI_v2
}  // namespace PacBio

#endif  // PBCOPPER_CLI_v2_HIDDENOPTIONMODE_H
