#ifndef PBCOPPER_DATA_ZMW_H
#define PBCOPPER_DATA_ZMW_H

#include <pbcopper/PbcopperConfig.h>

#include <cstdint>

namespace PacBio {
namespace Data {

typedef int32_t Zmw;

}  // namespace Data
}  // namespace PacBio

#endif  // PBCOPPER_DATA_ZMW_H
