#ifndef PBCOPPER_PBBOILERPLATEDISCLAIMER_H
#define PBCOPPER_PBBOILERPLATEDISCLAIMER_H

#define PB_BOILERPLATE_DISCLAIMER                                                              \
    "Copyright (C) 2004-2023     Pacific Biosciences of California, Inc.\nThis program comes " \
    "with ABSOLUTELY NO WARRANTY; it is intended for\nResearch Use Only and not for use in "   \
    "diagnostic procedures.\n"

#endif  // PBCOPPER_PBBOILERPLATEDISCLAIMER_H
